% Finite difference approximation function based on Rebrin et. al (1999),
% with model: Rearrange plasma-to-ISF diffusion process
% solved for the plasma glucoe (u) from ISF glucose (y)

function [output, error] = FDapprox(y, T, dt, t, ybga, tbga, varargin)

parsedArgs = parseInputVarArgs(varargin);

g = 1; % g is assumed to be 1 -> steady state gain
a = -1/T;
b = g/T;
uhat(1) = y(1); % Initialized as first measurement
uReg(1) = y(1); % Initialized as first measurement

Nreg = 250;     % Regression interval
Ns = 600;       % Ma smoothing interval

for i=2:length(y)
    tic;
    
    % Gp estimate
    uhat(i) = (1/(b*dt)-a/b)*y(i)+(-1/(b*dt))*y(i-1);
    
    %%%% MA smoothing of estimates
    if parsedArgs.useMAsmoothing && (mod(i,Ns) == 0)
        uhatS = smooth(uhat, Ns);
    else
        uhatS(i) = uhat(i);
    end
        
    %%%% WLS LR of estimates
    if parsedArgs.useRegression && (mod(i,Nreg) == 0)
        tint = t(i-Nreg+1:i); 
        yint = uhat(i-Nreg+1:i)';
        if i==Nreg
            yhat = linearRegression(tint, yint, 1, 1); 
            uReg(i-Nreg+1:i) = yhat;
        else
            yhat = linearRegression(tint-tReg(end), yint-uReg(end), 0, 1); 
            uReg(i-Nreg+1:i) = yhat+uReg(end);
        end
        tReg(i-Nreg+1:i) = t(i-Nreg+1:i);
    end
    
    time(i) = toc;
end

% Generate filter output
output.Gp = uhat;
output.t = t;
output.y = y;
output.ybga = ybga;
output.tbga = tbga;
output.maxItRunTime = max(time);

% Generate error output
error_filter = computeError(ybga, output.Gp, tbga, output.t);
error.filter = error_filter;

if parsedArgs.useMAsmoothing
    output.GpS = uhatS;
    error_mas = computeError(ybga, output.GpS, tbga, output.t(1:end-1));
    error.mas = error_mas;
end

if parsedArgs.useRegression
    output.GpReg = uReg;
    output.tReg = tReg;
    error_reg = computeError(ybga, output.GpReg, tbga, output.tReg);
    error.reg = error_reg;
end

if parsedArgs.plotResult
    % Plot Gp estimate
    figure(13)
    plot(t, uhat,'LineWidth',1.5,'DisplayName', '$\hat{G}_p$')
    hold on
    plot(t, y,'LineWidth',1.5, 'LineWidth',2,'DisplayName', 'CGM'), hold on
    hold on;
    plot(tbga, ybga, 'k.','MarkerSize',15,'DisplayName', 'BGA'), hold on
    grid on
    xlim([t(1) output.t(end)])
    ylim([-3 25])
    title('FD Approximation - $\hat{G}_p$','fontsize', 15,'interpreter', 'latex')
    legend('show', 'location', 'best', 'interpreter', 'latex','fontsize', 15)
    xlabel('Time [min]','fontsize', 15);
    ylabel('Glucose [mmol/L]','fontsize', 15);
    
    % Plot smoothed Gp estimate 
    if parsedArgs.useMAsmoothing 
        figure(23)
        plot(t, uhat, 'LineWidth',1.5,'DisplayName', '$\hat{G}_p$')
        hold on 
        plot(t, uhatS,'LineWidth',1.5, 'DisplayName', '$\hat{G}_{p,s}$')
        hold on
        plot(tbga, ybga, 'k.', 'Markersize', 15, 'DisplayName', 'BGA')
        legend('show','location','best', 'interpreter', 'latex','fontsize', 15)
        hold off
        grid on
        xlim([t(1) t(end)])
        ylim([-3 25])
        title('FD approximation - MA Smoother','fontsize', 15, 'interpreter', 'latex')
        xlabel('Time [min]','fontsize', 15);
        ylabel('Glucose [mmol/L]','fontsize', 15);
    end
    
    % Plot regressed Gp estimate
    if parsedArgs.useRegression
        figure(33)
        plot(output.t, output.Gp, 'LineWidth',1.5,'DisplayName', '$\hat{G}_p$')
        hold on
        plot(output.tReg, output.GpReg,'Linewidth', 1.5,'DisplayName', '$\hat{G}_{p,reg}$'); 
        hold on
        plot(tbga, ybga, 'k.', 'Markersize', 15, 'DisplayName', 'BGA')
        grid on
        legend('show','location','best', 'interpreter', 'latex','fontsize', 15)
        xlim([t(1) output.t(end)])
        ylim([-3 25])
        title('FD approximtaion - WLS LR','fontsize', 15, 'interpreter', 'latex')
        xlabel('Time [min]','fontsize', 15);
        ylabel('Glucose [mmol/L]','fontsize', 15);
    end
    
   
end

end

%%% Helper methods %%%
% Helper method to parse the arguments
function parsedArgs = parseInputVarArgs(varargs)
    %Set defaults
    parsedArgs.plotResult = 0;
    parsedArgs.useRegression = 0;
    parsedArgs.useMAsmoothing = 0;

    if mod(size(varargs,2),2)~=0
        error(['key/value arguments are not even:' num2str(size(varargs))])
    end
    for i = 1:2:size(varargs,2)
        if isfield(parsedArgs,varargs{i})
            parsedArgs.(varargs{i})=varargs{i+1};
        else
            error(['Invalid argument supplied:' num2str(varargs{i})])
        end
    end
end

