% M-Robust Kalman filter based on Z. M. Durovic et. al (1999),
% with model: rate-only + Steil-Rebrin model.
% Implementation based on https://github.com/milsto/robust-kalman 

function [output, error] = MrobustKF(y, t, T, a, b, dt, y_bga, t_bga, varargin)

parsedArgs = parseInputVarArgs(varargin);

% Handle time
if isdatetime(t)
    % Convert to relative time
    t.TimeZone='';
    startDateTime = t(1);
    t=convertToRelativeTime(t, startDateTime);
else
    %Assume relative time is passed in
    startDateTime = NaN;
end

% Set model
model = setModel(T, dt, a, 3);
if parsedArgs.useLowPassFilter
    model = augmentWithLPF(model, b, dt, 0);    
end
A = model.Phi;
H = model.H;
Q = model.Q;

% Storage
len = length(t);
Nstates = size(A, 1);
xhat_vec = nan(Nstates,len);        
P_vec = nan(Nstates,Nstates,len);

% Inititialize 
xhat_vec(:,1)= model.x0;
P_vec(:,:,1) = model.initCov;

Ns = 600;   % Regression interval
Nreg = 375; % MA smoothing interval

for i=1:len-1
    tic;
    
    %%% ISO standard R
    if y(i) < 5.55
        R = 0.172;
    else
        R = 0.0056*y(i)^2;
    end
    
    %%%% Low-pass filter measurements
    if parsedArgs.useLowPassFilter
        if i> 2
            ym(i)=(1-b)*ym(i-1)+b*y(i);
        else
            ym(i) = y(i);
        end
        z(i) = ym(i);
    else
        z(i) = y(i);
    end
    
    %%% Previous estimat 
    xprev = xhat_vec(:,i);
    Pprev = P_vec(:,:,i);
    
    %%% Prediction step / Time update
    [xbar, Pbar] = prediction(xprev,Pprev, A, Q);

    %%% Correction step / Measurement update   
    [xhat, P] = correction(xbar, Pbar, z(i), A, H, R);
    
    xhat_vec(:,i+1) = xhat;
    P_vec(:,:,i+1) = P;
    
    %%%% MA smoothing of estimates
    if parsedArgs.useMAsmoothing && (mod(i,Ns) == 0)
        xhatS = smooth(xhat_vec(1,1:i), Ns);
    else
        xhatS(i) = xhat(1);
    end
         
    %%%% WLS LR of estimates
    if parsedArgs.useRegression && (mod(i,Nreg) == 0) 
        tint = t(i-Nreg+1:i+1); 
        yint = xhat_vec(1,i-Nreg+1:i+1)';
        if i==Nreg
            yhat = linearRegression(tint, yint, 1, 1); 
            xReg(i-Nreg+1:i+1) = yhat;
        else
            yhat = linearRegression(tint-tReg(end), yint-xReg(end), 0, 0); 
            xReg(i-Nreg+1:i+1) = yhat+xReg(end);
        end
        tReg(i-Nreg+1:i+1) = t(i-Nreg+1:i+1);
    end
        
    time(i) = toc;
end

% Generate filter output
output.t = t;
output.Gp = xhat_vec(1,:);
output.dGp = xhat_vec(2,:);
output.Gisf = xhat_vec(3,:);
output.y = y;
output.ybga = y_bga;
output.tbga = t_bga;
output.maxItRunTime = max(time);

% Generate error output
error_filter = computeError(y_bga, output.Gp, t_bga, output.t);
error.filter = error_filter;

if parsedArgs.useLowPassFilter
    output.ylpf = z;
end
if parsedArgs.useMAsmoothing
    output.GpS = xhatS;
    error_mas = computeError(y_bga, output.GpS, t_bga, output.t(1:end-1));
    error.mas = error_mas;
end
if parsedArgs.useRegression
    output.GpReg = xReg;
    output.tReg = tReg;
    error_reg = computeError(y_bga, output.GpReg, t_bga, output.tReg);
    error.reg = error_reg;
end

% Get estimate standard deviations
for k = 1:length(t)
    output.Gp_sd(k) = sqrt(P_vec(1,1,k));
    output.dGp_sd(k) = sqrt(P_vec(2,2,k));
    output.Gisf_sd(k) = sqrt(P_vec(3,3,k));
end

if parsedArgs.plotResult

    % Plot Gp estimate
    figure(11)
    plot(output.t, output.Gp, 'LineWidth',1.5)
    hold on
    plot(t(1:end-1), z, 'LineWidth',1.5)
    hold on;
    plot(output.t, output.Gisf,'Linewidth', 1.5); 
    hold on
    plot(t_bga, y_bga, 'k.', 'Markersize', 15)
    plot(output.t, output.Gp+2*output.Gp_sd,'--','color','#A2142F','LineWidth',1)
    hold on
    plot(output.t, output.Gp-2*output.Gp_sd, '--','color','#A2142F', 'LineWidth',1)
    legend('$\hat{G}_p$','CGM','$\hat{G}_{isf}$' ,'BGA','$\pm$ SD $\hat{G}_{p}$','location','best','interpreter', 'latex','fontsize', 15)
    hold off
    grid on
    xlim([t(1) output.t(end)])
    title('M-Robust KF')
    xlabel('Time [min]')
    ylabel('Glucose [mmol/L]')

    % Plot regressed Gp estimate
    if parsedArgs.useRegression
        figure(11)
        plot(output.t, output.Gp, 'LineWidth',2)
        hold on
        plot(output.tReg, output.GpReg,'Linewidth', 1.5); 
        hold on
        plot(t_bga, y_bga, 'k.', 'Markersize', 15)
        grid on
        legend('$\hat{G}_p$','$\hat{G}_{p,reg}$','BGA','location','best','interpreter', 'latex','fontsize', 15)
        xlim([t(1) output.t(end)])
        title('M-Robust KF - WLS LR','interpreter', 'latex','fontsize', 15)
        xlabel('Time [min]');
        ylabel('Glucose [mmol/L]');
    end
    
    % Plot smoothed Gp estimate
    if parsedArgs.useMAsmoothing
        figure(12)
        plot(output.t, output.Gp, 'LineWidth',2)
        hold on
        plot(output.t(1:end-1), output.GpS,'Linewidth', 1.5); 
        hold on
        plot(t_bga, y_bga, 'k.', 'Markersize', 15)
        grid on
        legend('$\hat{G}_p$','$\hat{G}_{p,s}$','BGA','location','best','interpreter', 'latex','fontsize', 15)
        xlim([t(1) output.t(end)])
        title('M-Robust KF - MA Smoothing','interpreter', 'latex','fontsize', 15)
        xlabel('Time [min]');
        ylabel('Glucose [mmol/L]');
    end
    
    % Plot dGp estimate
    figure(21)
    plot(output.t, output.dGp, 'LineWidth',1.5)
    hold on
    plot(output.t, output.dGp+2*output.dGp_sd,'--','color','#A2142F','LineWidth',1)
    hold on
    plot(output.t, output.dGp-2*output.dGp_sd, '--','color','#A2142F', 'LineWidth',1)
    grid on
    legend('$d\hat{G}_{p}$','$\pm$ SD $d\hat{G}_{p}$','location','best','interpreter', 'latex','fontsize', 15)
    xlim([t(1) output.t(end)])
    title('M-Robust KF - $d\hat{G_p}$','interpreter', 'latex','fontsize', 15)
    
    % Plot Gisf estimate
    figure(81)
    plot(output.t, output.Gisf, 'LineWidth',1.5)
    hold on
    plot(output.t, output.Gisf+2*output.Gisf_sd,'--','color','#A2142F','LineWidth',1)
    hold on
    plot(output.t, output.Gisf-2*output.Gisf_sd, '--','color','#A2142F', 'LineWidth',1)
    grid on
    legend('$\hat{G}_{isf}$','$\pm$ SD $\hat{G}_{isf}$','location','best','interpreter', 'latex','fontsize', 15)
    xlim([t(1) output.t(end)])
    title('M-Robust KF - $\hat{G}_{isf}$','interpreter', 'latex','fontsize', 15)
end

end

%%% Helper methods %%%
function block_mat = findBlockMat(P, R)
nP = height(P);
mP = width(P);
nR = height(R);
mR = width(R);

block_mat = [P, zeros(nP,mR); zeros(nR, mP), R];
end

function crit = mEstimationCrit(beta, Y, X)
crit = 0;
for i=1:height(Y)
    crit = crit + HuberScore(Y(i,:)- X(i,:)*beta);
end
end

function score = HuberScore(z)
delta = 1.5; % Desierd efficiency at Gaussian model 
if abs(z) >= delta
    score = delta*abs(z) - power(delta, 2)/2;
else
    score = power(z, 2)/2;
end
end

function [xbar, Pbar]=prediction(xprev, Pprev, A, Q)
xbar = A*xprev;
Pbar = A*Pprev*A'+Q;
end

function [xhat, P] = correction(xbar, Pbar, z, F, H, R)
C = H*Pbar*H' + R;
K = (Pbar*H')/ C;

epsilon_cov = findBlockMat(Pbar, R); 
S = chol(epsilon_cov); % Cholesky factorization
Sinv = inv(S');

% Regression problem: Y = X*x + zeta
Y = Sinv*[xbar; z];
X = Sinv*[eye(height(xbar)); H];

% M-Estimat of x
%options = optimset('PlotFcns',@optimplotfval);
fun = @(xx)mEstimationCrit(xx, Y, X);
x = fminsearch(fun, xbar);%, options);

xhat = x;
P = (eye(height(F))-K*H)*Pbar;
res = z - H*xhat;
end

% Helper method to parse the arguments
function parsedArgs = parseInputVarArgs(varargs)
    %Set defaults
    parsedArgs.plotResult = 0;
    parsedArgs.useRegression = 0;
    parsedArgs.useAdaptiveCov = 0;
    parsedArgs.useLowPassFilter = 0;
    parsedArgs.useMAsmoothing = 0;

    if mod(size(varargs,2),2)~=0
        error(['key/value arguments are not even:' num2str(size(varargs))])
    end
    for i = 1:2:size(varargs,2)
        if isfield(parsedArgs,varargs{i})
            parsedArgs.(varargs{i})=varargs{i+1};
        else
            error(['Invalid argument supplied:' num2str(varargs{i})])
        end
    end
end

