clc
clear all
close all

%% Set model 
T  = 6;       % Time constant for diffusion from blood plasma to ISF [min]
a  = -0.0018; % Rate of decay [-]
dt = 1.2/60;  % Sampling rate [min]
t  = 1:dt:5;  % Simulation window
model = setROModel(T,a,dt); % Rate-only + Steil-Rebrin model

%% Sys
L = length(t);
F = model.phi;
G = model.G;
H = model.H;
x0 = model.x0;
P0 = model.P0;
sigma_w = model.sigma_w;
sigma_v = model.sigma_v;
x0_kf = model.x0_kf;
Q = model.Q;
R = model.R;

%% Storing
len = length(L);
Nstates = size(F, 1);
x_vec = nan(Nstates,len);
y_vec = nan(1, len);

%% Inititialize
x_vec(:,1) = x0;
y_vec(1) = H*x0;
xhat_vecMest(:,1)= x0_kf;
P_vecMest(:,:,1) = P0;
xhat_vecKF(:,1)= x0_kf;
P_vecKF(:,:,1) = P0;

%% Loop
for i=1:L-1 
    %%% Simulated system
    q = sigma_w*(0 + (0+1)*randn(1,1));
    
    if unifrnd(0,1) > 0.9
        rare_event = 1;
    else
        rare_event = 0;
    end
    r = sigma_v*randn(1,1)+randsample(-6:6,1)*unifrnd(1,1.5)*rare_event;
    
    x_vec(:,i+1) = F*x_vec(:,i)+G*q;
    y_vec(i) = H*x_vec(:,i)+ r;
    
    %%%% Filters
    %%% Previous estimat 
    xprev1 = xhat_vecMest(:,i);
    Pprev1 = P_vecMest(:,:,i);
    
    xprev2 = xhat_vecKF(:,i);
    Pprev2 = P_vecKF(:,:,i);
    
    %%% Correction step / Measurement update   
    [x1, P1, res] = Mest(xprev1, Pprev1, y_vec(i), F, H, Q, R);
    [x2, P2] = KF(xprev2, Pprev2, y_vec(i), F, H, Q, R); 
    
    xhat_vecMest(:,i+1) = x1;
    P_vecMest(:,:,i+1) = P1;
    
    xhat_vecKF(:,i+1) = x2;
    P_vecKF(:,:,i+1) = P2;
   
end

% Figures
fig = figure(1);
% M-robust KF estimate
subplot(312)
plot(t(1:end-1), y_vec,'DisplayName', 'CGM')
hold on
plot(t, xhat_vecMest(1,:),'color', '#D95319','Linewidth', 2,'DisplayName', '$\hat{G}_{isf}$')
grid on
xlim([t(1) t(end)])
h1 = legend('show');
set( h1, 'Interpreter','latex','fontsize', 15,'location', 'best')
title('M-Robust KF','fontsize', 15)

% KF estimate
subplot(313)
plot(t(1:end-1), y_vec,'DisplayName', 'CGM')
hold on
plot(t, xhat_vecKF(1,:), 'color', '#7E2F8E','Linewidth', 2,'DisplayName', '$\hat{G}_{isf}$')
grid on
xlim([t(1) t(end)])
h1 = legend('show');
set( h1, 'Interpreter','latex','fontsize', 15,'location', 'best')
title('KF','fontsize', 15)

% True value
subplot(311)
plot(t(1:end-1), y_vec,'DisplayName', 'CGM')
hold on
plot(t, x_vec(1,:), 'color', '#77AC30','Linewidth', 2,'DisplayName', '$G_{isf}$')
grid on
xlim([t(1) t(end)])
h1 = legend('show');
set( h1, 'Interpreter','latex','fontsize', 15,'location', 'best')
title('Real system','fontsize', 15)

%% Helper methods %%

% Rate only model
function model = setROModel(T,a,dt)
model.F = [-1/T 1/T 0; 0 0 1; 0 0 a];
model.G = [0; 0; 1];
model.H = [1 0 0];
model.x0 = [5; 5; 0];
model.P0 = diag([1 1 1]);

if ~rank(obsv(model.F,model.H))==size(model.F,1)  % Checking observability
    disp('System is not observable with only cgm measurements')
end

model.sigma_w = 1;
model.sigma_v = 0.5;

model.x0_kf = [5; 5; 5];

model.Q = (model.G*model.G')*model.sigma_w^2;
model.R = model.sigma_v^2;

model.phi=expm(model.F*dt);
end

function block_mat = findBlockMat(P, R)
nP = height(P);
mP = width(P);
nR = height(R);
mR = width(R);

block_mat = [P, zeros(nP,mR); zeros(nR, mP), R];
end

function crit = mEstimationCrit(beta, Y, X)
crit = 0;
for i=1:height(Y)
    crit = crit + HuberScore(Y(i,:)- X(i,:)*beta);
end
end

function score = HuberScore(z)
delta = 1.5;
if abs(z) >= delta
    score = delta*abs(z) - power(delta, 2)/2;
else
    score = power(z, 2)/2;
end
end

% M-robust KF
function [xbar, Pbar] = prediction(xprev, Pprev, F, Q)
xbar = F*xprev;
Pbar = F*Pprev*F'+Q;
end

function [xhat, P, res] = correction(xbar, Pbar, z, F, H, R)

C = H*Pbar*H' + R;
K = (Pbar*H')/ C;

epsilon_cov = findBlockMat(Pbar, R); 
S = chol(epsilon_cov); % Cholesky factorization
Sinv = inv(S');

% Regression problem: Y = X*x + zeta
Y = Sinv*[xbar; z];
X = Sinv*[eye(height(xbar)); H];

% M-Estimat of x
% options = optimset('PlotFcns',@optimplotfval);
fun = @(xx)mEstimationCrit(xx, Y, X);
x = fminsearch(fun, xbar); %, options);

xhat = x;
P = (eye(height(F))-K*H)*Pbar;
res = z - H*xhat;
end

% KF
function [xhat, P] = correction2(xbar, Pbar, z, F, H, R)
S = H*Pbar*H'+R;
K = (Pbar*H')/S;
xhat = xbar + K*(z-H*xbar);
P = Pbar-K*H*Pbar;
end

%%%%% Compact %%%%%
function [x, P, res] = Mest(xprev, Pprev, z, F, H, Q, R)
[xbar, Pbar] = prediction(xprev,Pprev, F, Q);
[x, P, res] = correction(xbar, Pbar, z, F, H, R);
end

function [x, P] = KF(xprev, Pprev, z, F, H, Q, R)
[xbar, Pbar] = prediction(xprev,Pprev, F, Q);
[x, P] = correction2(xbar, Pbar, z, F, H, R);
end


