% Robust Two Stage Kalman Filter with Unknown inputs based on Chien-Shu
% Hsieh (2000), with model: Steil-Rebrin with BG level as unknown input.

function [output, error] = robustTwoStageKFui(y, t, T, a, dt, t_bga, y_bga, varargin)

parsedArgs = parseInputVarArgs(varargin);

len = length(t);

% Handle time
if isdatetime(t)
    % Convert to relative time
    t.TimeZone='';
    startDateTime = t(1);
    t=convertToRelativeTime(t, startDateTime);
else
    %Assume relative time is passed in
    startDateTime = NaN;
end

% Set model
model=setModel(T, dt, a, 4);
A = model.Phi;
H = model.H;
Q = model.Q;
R = model.R;
E = model.Eps;

%Storing
Nstates = size(A, 1);
xhat = zeros(Nstates,len);        
d = zeros(1, len);  
Px = zeros(Nstates,Nstates, len);
Pd = zeros(Nstates, len);

% Initialize 
Px(:,:,1) = model.initCov;
xhat(:,1) = model.x0;

I = eye(size(A,1)); % Identity matrix

for i=1:len-1
    tic;
    
    %%% ISO standard R
    if y(i) < 5.55
        R = 0.172;
    else
        R = 0.0056*y(i)^2;
    end

    xhatPrev = xhat(:,i);
    PxPrev = Px(:,:,i);
    
    %%% Stage 1: 
    %%%% Prediction x
    xbar =  A *xhatPrev;   
    Px_bar = A*PxPrev*A'+Q;

    C = H*Px_bar*H'+R;

    Kx = Px_bar*H'*inv(C);

    %%%% Correcting bar estimates
    Px_hat_temp = (I - Kx*H)*Px_bar;
    xhat_temp = xbar + Kx*(y(i) - H *xbar);

    %%% Stage 2: 
    %%%% Predicting d
    Pd(i+1) = inv(E'*H'*inv(C)*H*E);
    Kd = Pd(i+1)*E'*H'*inv(C);
    d(i+1) = Kd*(y(i)-H*xbar);
    Kdlist(i) = Kd;

    %%%% Blending matrix
    V = (I-Kx*H)*E;

    %%%% Correction x
    xhat(:,i+1) = xhat_temp + V*d(i+1);
    Px(:,:,i+1) = Px_hat_temp+ V*Pd(i+1)*V';

   time(i) = toc;
end

% Generate output object
output.t = t;
output.Gisf = xhat(1,:);
output.Gp = d;
output.y = y;
output.ybga = y_bga;
output.tbga = t_bga;
output.maxItRunTime = max(time);

% Generate error output
error = computeError(y_bga, output.Gp, t_bga, output.t);

for k = 1:length(t)
    output.Gp_sd(k) = sqrt(Pd(k));
    output.Gisf_sd(k) = sqrt(Px(k));
end

if parsedArgs.plotResult
    % Plot Gp estimate
    figure(10)
    plot(output.t, output.Gp, 'LineWidth',1.5)
    hold on
    plot(t, y, 'LineWidth',1.5)
    hold on;
    plot(output.t, output.Gisf,'Linewidth', 1.5); 
    hold on
    plot(t_bga, y_bga, 'k.', 'Markersize', 15)
    plot(output.t, output.Gp+2*output.Gp_sd,'--','color','#A2142F','LineWidth',1)
    hold on
    plot(output.t, output.Gp-2*output.Gp_sd, '--','color','#A2142F', 'LineWidth',1)
    legend('$\hat{G}_p$','CGM','$\hat{G}_{isf}$' ,'BGA','$\pm$ SD $\hat{G}_{p}$','location','best','interpreter', 'latex','fontsize', 15)
    hold off
    grid on
    xlim([t(1) output.t(end)])
    title('RTSKF UI - $\hat{G}_p$','fontsize', 15,'interpreter', 'latex')
    xlabel('Time [min]');
    ylabel('Glucose [mmol/L]');
    
    % Plot Gisf estimate
    figure(22)
    plot(output.t, output.Gisf,'Linewidth', 1.5); 
    hold on
    plot(t, y, 'LineWidth',1.5)
    hold on;
    plot(t_bga, y_bga, 'k.', 'Markersize', 15)
    plot(output.t, output.Gisf+2*output.Gisf_sd,'--','color','#A2142F','LineWidth',1)
    hold on
    plot(output.t, output.Gisf-2*output.Gisf_sd, '--','color','#A2142F', 'LineWidth',1)
    legend('$\hat{G}_{isf}$','CGM' ,'BGA','$\pm$ SD $\hat{G}_{isf}$','location','best','interpreter', 'latex','fontsize', 15)
    hold off
    grid on
    xlim([t(1) output.t(end)])
    title('RTSKF UI -$\hat{G}_{isf}$','interpreter', 'latex','fontsize', 15)
    xlabel('Time [min]');
    ylabel('Glucose [mmol/L]');
    
end
end

% Helper method to parse the arguments
function parsedArgs = parseInputVarArgs(varargs)
    %Set defaults
    parsedArgs.plotResult = 0;

    if mod(size(varargs,2),2)~=0
        error(['key/value arguments are not even:' num2str(size(varargs))])
    end
    for i = 1:2:size(varargs,2)
        if isfield(parsedArgs,varargs{i})
            parsedArgs.(varargs{i})=varargs{i+1};
        else
            error(['Invalid argument supplied:' num2str(varargs{i})])
        end
    end
end
