% Kalman filter based on "Optimal State Estimation" - Dan Simon (2006),
% with model: rate-only + Steil-Rebrin model.

function [output, error] = standardKF(y, t, Tisf, a, b, dt, y_bga, t_bga, varargin)

parsedArgs = parseInputVarArgs(varargin);

% Handle time
if isdatetime(t)
    % Convert to relative time
    t.TimeZone='';
    startDateTime = t(1);
    t=convertToRelativeTime(t, startDateTime);
else
    % Assume relative time is passed in
    startDateTime = NaN;
end

% Set model
model = setModel(Tisf, dt, a, 3);
if parsedArgs.useLowPassFilter
    model = augmentWithLPF(model, b, dt, 0);    
end
A = model.Phi;
H = model.H;
Q = model.Q;
R = model.R;

% Storing
Nstates = size(A, 1);
xhat = nan(Nstates,length(t));        
P = nan(Nstates,Nstates,length(t));

% Initialization
xhat(:,1) = model.x0;
P(:,:,1) = model.initCov;
ym(1) = 0;
z(1) = 0;

Ns = 600;   % Regression interval
Nreg = 375; % MA smoothing interval

for i=1:length(t)-1
    tic;
    
    %%% ISO standard R
    if y(i) < 5.55
        R = 0.172;
    else
        R = 0.0056*y(i)^2;
    end
    
    %%%% Low-pass filter measurements
    if parsedArgs.useLowPassFilter
        if i> 2
            ym(i)=(1-b)*ym(i-1)+b*y(i);
        else
            ym(i) = y(i);
        end
        z(i) = ym(i);
    else
        z(i) = y(i);
    end
    
    %%%% Prediction step
    xbar = A*xhat(:,i);
    P_bar = A*P(:,:,i)*A'+ Q;

    %%%% Correction step
    innov = (z(i)-H*xbar);
    S = H*P_bar*H'+R;
    K = (P_bar*H')/S;
    xhat(:,i+1) = xbar + K*innov;
    P(:,:,i+1) = P_bar-K*H*P_bar;

    %%%% MA smoothing of estimates
    if parsedArgs.useMAsmoothing && (mod(i,Ns) == 0)
        xhatS = smooth(xhat(1,1:i), Ns);
    else
        xhatS(i) = xhat(1,i);
    end
    
    %%%% WLS LR of estimates
    if parsedArgs.useRegression && (mod(i,Nreg) == 0)       
        tint = t(i-Nreg+1:i+1); 
        yint = xhat(1,i-Nreg+1:i+1)';
        if i==Nreg
            yhat = linearRegression(tint, yint, 1, 1); 
            xReg(i-Nreg+1:i+1) = yhat;
        else
            yhat = linearRegression(tint-tReg(end), yint-xReg(end), 0, 0); 
            xReg(i-Nreg+1:i+1) = yhat+xReg(end);
        end
        tReg(i-Nreg+1:i+1) = t(i-Nreg+1:i+1);
    end
    
    time(i) = toc;
end

% Generate filter output
output.t = t;
output.Gp = xhat(1,:);
output.dGp = xhat(2,:);
output.Gisf = xhat(3,:);
output.y = y;
output.ybga = y_bga;
output.tbga = t_bga;
output.maxItRunTime = max(time);

% Generate error output
error_filter = computeError(y_bga, output.Gp, t_bga, output.t);
error.filter = error_filter;

if parsedArgs.useLowPassFilter
    output.ylpf = z;
end
if parsedArgs.useMAsmoothing
    output.GpS = xhatS;
    error_mas = computeError(y_bga, output.GpS, t_bga, output.t(1:end-1));
    error.mas = error_mas;
end
if parsedArgs.useRegression
    output.GpReg = xReg;
    output.tReg = tReg;
    error_reg = computeError(y_bga, output.GpReg, t_bga, output.tReg);
    error.reg = error_reg;
end

% Get estimate standard deviations
for k = 1:length(t)
    output.Gp_sd(k) = sqrt(P(1,1,k));
    output.dGp_sd(k) = sqrt(P(2,2,k));
    output.Gisf_sd(k) = sqrt(P(3,3,k));
end

if parsedArgs.plotResult
   % Plot Gp estimate
    fig6 = figure(6);
    plot(output.t, output.Gp, 'LineWidth',1.5)
    hold on
    plot(t(1:end-1), z, 'LineWidth',1.5)
    hold on;
    plot(output.t, output.Gisf,'Linewidth', 1.5); 
    hold on
    plot(t_bga, y_bga, 'k.', 'Markersize', 15)
    plot(output.t, output.Gp+2*output.Gp_sd,'--','color','#A2142F','LineWidth',1)
    hold on
    plot(output.t, output.Gp-2*output.Gp_sd, '--','color','#A2142F', 'LineWidth',1)
    legend('$\hat{G}_p$','CGM','$\hat{G}_{isf}$' ,'BGA','$\pm$ SD $\hat{G}_{p}$','location','northeast','interpreter', 'latex','fontsize', 15)
    hold off
    grid on
    xlim([t(1) output.t(end)])
    title('KF - $\hat{G}_p$','fontsize', 15,'interpreter', 'latex')
    xlabel('Time [min]');
    ylabel('Glucose [mmol/L]');
 
    % Plot regressed Gp estimate
    if parsedArgs.useRegression
        figure(5)
        plot(output.t, output.Gp, 'LineWidth',2)
        hold on
        plot(output.tReg, output.GpReg,'Linewidth', 1.5); 
        hold on
        plot(t_bga, y_bga, 'k.', 'Markersize', 15)
        grid on
        legend('$\hat{G}_p$','$\hat{G}_{p,reg}$','BGA','location','northeast','interpreter', 'latex','fontsize', 15)
        xlim([t(1) output.t(end)])
        ylim([2 14])
        title('KF - WLS LR', 'interpreter', 'latex','fontsize', 15)
        xlabel('Time [min]');
        ylabel('Glucose [mmol/L]');
    end
    
    % Plot smoothed Gp estimate
    if parsedArgs.useMAsmoothing
        figure(7)
        plot(output.t, output.Gp, 'LineWidth',2)
        hold on
        plot(output.t(1:end-1), output.GpS,'Linewidth', 1.5); 
        hold on
        plot(t_bga, y_bga, 'k.', 'Markersize', 15)
        grid on
        legend('$\hat{G}_p$','$\hat{G}_{p,s}$','BGA','location','northeast','interpreter', 'latex','fontsize', 15)
        xlim([t(1) output.t(end)])
        ylim([2 14])
        title('KF - MA Smoothing', 'interpreter', 'latex','fontsize', 15)
        xlabel('Time [min]');
        ylabel('Glucose [mmol/L]');
    end
    
    % Plot dGp estimate
    figure(8)
    plot(output.t, output.dGp, 'LineWidth',1.5)
    hold on
    plot(output.t, output.dGp+2*output.dGp_sd,'--','color','#A2142F','LineWidth',1)
    hold on
    plot(output.t, output.dGp-2*output.dGp_sd, '--','color','#A2142F', 'LineWidth',1)
    grid on
    legend('$d\hat{G}_{p}$','$\pm$ SD $d\hat{G}_{p}$','location','best','interpreter', 'latex','fontsize', 15)
    xlim([t(1) output.t(end)])
    title('KF - $d\hat{G_p}$','interpreter', 'latex','fontsize', 15)
    
    % Plot Gisf estimate
    figure(9)
    plot(output.t, output.Gisf, 'LineWidth',1.5)
    hold on
    plot(output.t, output.Gisf+2*output.Gisf_sd,'--','color','#A2142F','LineWidth',1)
    hold on
    plot(output.t, output.Gisf-2*output.Gisf_sd, '--','color','#A2142F', 'LineWidth',1)
    grid on
    legend('$\hat{G}_{isf}$','$\pm$ SD $\hat{G}_{isf}$','location','best','interpreter', 'latex','fontsize', 15)
    xlim([t(1) output.t(end)])
    title('KF - $\hat{G}_{isf}$','interpreter', 'latex','fontsize', 15)
end
end


% Helper method to parse the arguments
function parsedArgs = parseInputVarArgs(varargs)
    %Set defaults
    parsedArgs.plotResult = 0;
    parsedArgs.useRegression = 0;
    parsedArgs.useLowPassFilter = 0;
    parsedArgs.useMAsmoothing = 0;

    if mod(size(varargs,2),2)~=0
        error(['key/value arguments are not even:' num2str(size(varargs))])
    end
    for i = 1:2:size(varargs,2)
        if isfield(parsedArgs,varargs{i})
            parsedArgs.(varargs{i})=varargs{i+1};
        else
            error(['Invalid argument supplied:' num2str(varargs{i})])
        end
    end
end

